import os
import json
import yaml
from subprocess import CalledProcessError
import tempfile
import pluggy

from hash import target_hookimpl as hookimpl

from hash import errors, resources

hookspec = pluggy.HookspecMarker("hash-targets")


class TargetSpec:
    @hookspec
    def action(name: str, config: dict):
        pass

    @hookspec
    def _init(config: dict):
        pass


class Target(object):
    def _init(self, config: dict) -> None:
        self.name = config.get("name")
        if self.name is None:
            raise errors.TargetConfigError("No name specified for the target")
        self.kind = config.get("kind")
        if self.kind is None:
            raise errors.TargetConfigError("No kind specified for the target")
        self.spec = config.get("spec", {})


class FakeTarget(Target):
    @hookimpl
    def init(self, config: dict, space) -> None:
        self._init(config)
        if self.kind != "FakeTarget":
            raise errors.TargetConfigError(
                f"This target is not of kind Fake, it is of kind {self.kind}"
            )
        self.__space = space

    @hookimpl
    def action(name: str, config: dict, artifacts):
        pass


class K8STarget(Target):
    @hookimpl
    def init(self, config: dict, space) -> None:
        self._init(config)
        if self.kind != "K8STarget":
            raise errors.TargetConfigError(
                f"This target is not of kind K8S, it is of kind {self.kind}"
            )
        self.__space = space

    @hookimpl
    def action(self, name: str, config: dict):
        if name != "deploy" and name != "test" and name != "create_namespace":
            raise errors.TargetActionError(
                f"Cannot run action {name} with target of type K8S"
            )
        kubectl_binary = config.get("kubectl", "kubectl")
        context = self.spec.get("context")
        k8s_certificate_ca = self.spec.get("k8s_certificate_ca")
        if k8s_certificate_ca:
            try:
                open(k8s_certificate_ca, "r")
            except FileNotFoundError:
                t = k8s_certificate_ca
                k8s_certificate_ca = tempfile.mkstemp(text=True)
                with open(k8s_certificate_ca[1], "w") as f:
                    f.write(t)
                k8s_certificate_ca = k8s_certificate_ca[1]
            except OSError as e:
                if e.errno == 36:  # File name too long
                    t = k8s_certificate_ca
                    k8s_certificate_ca = tempfile.mkstemp(text=True)
                    with open(k8s_certificate_ca[1], "w") as f:
                        f.write(t)
                    k8s_certificate_ca = k8s_certificate_ca[1]
                else:
                    pass  # TODO: raise error
        k8s_certificate = self.spec.get("k8s_certificate")
        if k8s_certificate:
            try:
                open(k8s_certificate, "r")
            except FileNotFoundError:
                t = k8s_certificate
                k8s_certificate = tempfile.mkstemp(text=True)
                with open(k8s_certificate[1], "w") as f:
                    f.write(t)
                k8s_certificate = k8s_certificate[1]
            except OSError as e:
                if e.errno == 36:
                    t = k8s_certificate
                    k8s_certificate = tempfile.mkstemp(text=True)
                    with open(k8s_certificate[1], "w") as f:
                        f.write(t)
                    k8s_certificate = k8s_certificate[1]
                else:
                    pass
        k8s_client_key = self.spec.get("k8s_client_key")
        if k8s_client_key:
            try:
                open(k8s_client_key, "r")
            except FileNotFoundError:
                t = k8s_client_key
                k8s_client_key = tempfile.mkstemp(text=True)
                with open(k8s_client_key[1], "w") as f:
                    f.write(t)
                k8s_client_key = k8s_client_key[1]
            except OSError as e:
                if e.errno == 36:
                    t = k8s_client_key
                    k8s_client_key = tempfile.mkstemp(text=True)
                    with open(k8s_client_key[1], "w") as f:
                        f.write(t)
                    k8s_client_key = k8s_client_key[1]
                else:
                    pass
        k8s_kube_config = self.spec.get("k8s_kube_config")
        if k8s_kube_config:
            try:
                open(k8s_kube_config, "r")
            except FileNotFoundError:
                t = k8s_kube_config
                k8s_kube_config = tempfile.mkstemp(text=True)
                with open(k8s_kube_config[1], "w") as f:
                    f.write(t)
                k8s_kube_config = k8s_kube_config[1]
            except OSError as e:
                if e.errno == 36:
                    t = k8s_kube_config
                    k8s_kube_config = tempfile.mkstemp(text=True)
                    with open(k8s_kube_config[1], "w") as f:
                        f.write(t)
                    k8s_kube_config = k8s_kube_config[1]
                else:
                    pass
        server = self.spec.get("server")
        token = self.spec.get("token")
        config_kind = config.get("kind", "Kustomize")
        kubectl_args = ""
        path = config.get("path")
        if path is None:
            raise errors.ResourcePublishError(
                "You need to specify path for command execution"
            )
        use_gcloud = False
        if self.spec.get("use_gcloud") == True:
            use_gcloud = True
            cluster_name = self.spec.get("cluster_name")
            if cluster_name is None:
                raise errors.ActionError(
                    "use_gcloud is set to True but cluster_name is not set"
                )
            project_name = self.spec.get("project_name")
            if project_name is None:
                raise errors.ActionError(
                    "use_gcloud is set to True but project_name is not set"
                )
            cluster_region = self.spec.get("cluster_region")
            if cluster_region is None:
                cluster_zone = self.spec.get("cluster_zone")
                if cluster_zone is None:
                    raise errors.ActionError(
                        "neither cluster_region nor cluster_zone is set"
                    )
                try:
                    gcloud_command = f"gcloud container clusters get-credentials {cluster_name} --zone {cluster_zone} --project {project_name}"
                    os.environ["KUBECONFIG"] = os.path.join(
                        self.__space.get_hash_dir(),
                        f"{cluster_name}-{project_name}-config",
                    )
                    resources.Resource.execute(gcloud_command, path)
                except CalledProcessError as e:
                    raise errors.ActionError(e)
            else:
                try:
                    gcloud_command = f"gcloud container clusters get-credentials {cluster_name} --region {cluster_region} --project {project_name}"
                    os.environ["KUBECONFIG"] = os.path.join(
                        self.__space.get_hash_dir(),
                        f"{cluster_name}-{project_name}-config",
                    )
                    resources.Resource.execute(gcloud_command, path)
                except CalledProcessError as e:
                    raise errors.ActionError(e)
        use_az = False
        if self.spec.get("use_az") == True:
            use_az = True
            cluster_name = self.spec.get("cluster_name")
            if cluster_name is None:
                raise errors.ActionError(
                    "use_az is set to True but cluster_name is not set"
                )
            subscription_id = self.spec.get("subscription_id")
            if subscription_id is None:
                raise errors.ActionError(
                    "use_az is set to True but subscription_id is not set"
                )
            resource_group = self.spec.get("resource_group")
            if resource_group is None:
                raise errors.ActionError(
                    "use_az is set to True but resource_group is not set"
                )
            try:
                az_command = f"az aks get-credentials --name {cluster_name} --resource-group {resource_group} --subscription {subscription_id} --overwrite-existing"
                os.environ["KUBECONFIG"] = os.path.join(
                    self.__space.get_hash_dir(),
                    f"{cluster_name}-{resource_group}-{subscription_id}-config",
                )
                resources.Resource.execute(az_command, path)
            except CalledProcessError as e:
                raise errors.ActionError(e)
        if not use_gcloud and not use_az:
            if context:
                kubectl_args = f" --context={context} "
            else:
                if k8s_certificate_ca and k8s_certificate and k8s_client_key:
                    kubectl_args = f" --certificate-authority={k8s_certificate_ca} --client-certificate={k8s_certificate} --client-key={k8s_client_key} "
                if server:
                    kubectl_args += f" --server={server} "
                if token:
                    kubectl_args += f" --token={token} "
                if k8s_kube_config:
                    kubectl_args += f" --kubeconfig={k8s_kube_config} "
        if config_kind not in ["Kustomize", "DockerImage"]:
            raise errors.ResourceDeployError(
                "Config kind must be either 'Kustomize' or 'DockerImage'"
            )
        port = config.get("port")
        if port:
            kubectl_args += f"--port={port}"
        service_account = config.get("service_account")
        if service_account:
            kubectl_args += f"--serviceaccount={service_account}"
        if config_kind == "Kustomize" and name != "create_namespace":
            manifests_path = config.get("manifests_path")
            if manifests_path is None:
                raise errors.ResourceError(f"No manifests path supplied for K8S target")
            if name == "deploy":
                kubectl_command = (
                    f"{kubectl_binary} apply {kubectl_args} -f {manifests_path}"
                )
            elif name == "test" and config.get("diff") is not True:
                kubectl_command = f"{kubectl_binary} apply {kubectl_args} --dry-run=server -f {manifests_path}"
            elif name == "test" and config.get("diff") is True:
                kubectl_command = (
                    f"{kubectl_binary} diff {kubectl_args} -f {manifests_path}"
                )
        elif config_kind == "DockerImage":
            image_url = config.get("image_url")
            if image_url is None:
                raise errors.ResourceError("No image URL supplied for K8S target")
            pod_name = config.get("pod_name", name)
            kubectl_command = (
                f"{kubectl_binary} run {kubectl_args} --image={image_url} {pod_name}"
            )
        if name == "create_namespace":
            namespace = config.get("namespace")
            if namespace is None:
                raise errors.ResourceError(
                    "k8s target: create_namespace is reqested but no namespace name is provided"
                )
            kubectl_command = (
                f"{kubectl_binary} create namespace {kubectl_args} {namespace}"
            )
        try:
            resources.Resource.execute(kubectl_command, path)
        except CalledProcessError as e:
            if config.get("diff") is True and e.returncode == 1:
                return True
            if name == "create_namespace":
                raise e
            raise errors.ResourceError(e.stderr.decode("UTF-8"))
        finally:
            if k8s_certificate_ca:
                try:
                    os.remove(k8s_certificate_ca)
                except Exception:
                    pass
            if k8s_certificate:
                try:
                    os.remove(k8s_certificate)
                except Exception:
                    pass
            if k8s_client_key:
                try:
                    os.remove(k8s_client_key)
                except Exception:
                    pass
            if k8s_kube_config:
                try:
                    os.remove(k8s_kube_config)
                except Exception:
                    pass


class DockerRegistryTarget(Target):
    @hookimpl
    def init(self, config: dict, space) -> None:
        self._init(config)
        if self.kind != "DockerRegistryTarget":
            raise errors.TargetConfigError(
                f"This target is not of kind DockerRegistry, it is of kind {self.kind}"
            )
        self.__space = space

    @hookimpl
    def action(self, name: str, config: dict):
        if name != "publish":
            raise errors.TargetActionError(
                f"Cannot run action {name} with target of type Docker Registry"
            )
        registry_url = self.spec.get("registry_url")
        if registry_url is None:
            raise errors.ResourcePublishError("No registry_url in target")
        docker_config_dir = None
        credHelpers = self.spec.get("credHelpers")
        if credHelpers:
            docker_config_dir = tempfile.mkdtemp()
            with open(os.path.join(docker_config_dir, "config.json"), "w") as f:
                out = {"credHelpers": credHelpers}
                json.dump(out, f)
        docker_config = self.spec.get("docker_config")
        if docker_config:
            docker_config_dir = tempfile.mkdtemp()
            with open(os.path.join(docker_config_dir, "config.json"), "w") as f:
                f.write(docker_config)
        config_kind = config.get("kind", "DockerFile")
        if config_kind not in ["DockerFile", "DockerImage"]:
            raise errors.ResourcePublishError(
                "target connfig kind must be either 'DockerFile' or 'DockerImage'"
            )
        service_account = self.spec.get("service_account")
        login_command = None
        if service_account:
            registry_host = registry_url.split("/")[0]
            login_command = f"gcloud auth print-access-token --impersonate-service-account {service_account} | docker login -u oauth2accesstoken --password-stdin https://{registry_host}"
        use_az = self.spec.get("use_az")
        if use_az == True:
            registry_name = self.spec.get("registry_name")
            if registry_name is None:
                raise errors.ActionError("use_az is True but registry_name is not set")
            subscription_id = self.spec.get("subscription_id")
            if subscription_id is None:
                raise errors.ActionError(
                    "use_az is True but subscription_id is not set"
                )
            login_command = (
                f"az acr login --subscription {subscription_id} --name {registry_name}"
            )
        if config_kind == "DockerFile":
            image_name = config.get("image_name")
            if image_name is None:
                raise errors.ResourcePublishError("You need to specify image name")
            image_file = config.get("image_file")
            docker_file = config.get("docker_file", "Dockerfile")
            docker_file_path = config.get("docker_file_path", ".")
            image_url = f"{registry_url}/{image_name}"
            tag_command = None
            if docker_config_dir:
                if image_file:
                    build_command = f"docker image load -i {image_file}"
                    tag_command = f"docker tag {image_name} {image_url}"
                else:
                    build_command = f"docker --config={docker_config_dir} build -t {image_url} {docker_file_path} -f {docker_file}"
                publish_command = (
                    f"docker --config={docker_config_dir} push {image_url}"
                )
            else:
                if image_file:
                    build_command = f"docker image load -i {image_file}"
                    tag_command = f"docker tag {image_name} {image_url}"
                else:
                    build_command = f"docker build -t {image_url} {docker_file_path} -f {docker_file}"
                publish_command = f"docker push {image_url}"
            path = config.get("path")
            if path is None:
                raise errors.ResourcePublishError(
                    "You need to specify path for command execution"
                )
            try:
                if login_command:
                    resources.Resource.execute(login_command, path)
                resources.Resource.execute(build_command, path)
                if tag_command:
                    resources.Resource.execute(tag_command, path)
                resources.Resource.execute(publish_command, path)
                return image_url
            except CalledProcessError as e:
                raise errors.ResourcePublishError(e.stderr.decode("UTF-8"))


class DOFunctionTarget(Target):
    def __verify_function(
        self, package_name: str, function_name: str, config: dict
    ) -> bool:
        packages = config.get("packages", [])
        for package in packages:
            if package.get("name") == package_name:
                functions = package.get("functions", [])
                for function in functions:
                    if function.get("name") == function_name:
                        return True
        return False

    @hookimpl
    def init(self, config: dict, space) -> None:
        self._init(config)
        if self.kind != "DOFunctionTarget":
            raise errors.TargetConfigError(
                f"This target is not of kind DOFunction, it is of kind {self.kind}"
            )
        self.__space = space

    @hookimpl
    def action(self, name: str, config: dict):
        if name != "deploy":
            raise errors.ResourceError(
                f"DOFunction target can only run deploy actions not {name}"
            )
        region = self.spec.get("region")
        if region is None:
            raise errors.ResourceError("No region found in DOFunction target specs")
        project_dir = config.get("project_dir")
        if project_dir is None:
            raise errors.ResourceError(
                "No project_dir provided in config for DOFunction target"
            )
        if not os.path.exists(os.path.join(project_dir, "project.yml")):
            raise errors.ResourceError(f"No project.yml file found in {project_dir}")
        package_name = config.get("package_name")
        if package_name is None:
            raise errors.ResourceError(
                "No package_name is provided for DOFunction target"
            )
        function_name = config.get("function_name")
        if function_name is None:
            raise errors.ResourceError(
                "No function_name is provided for DOFunction target"
            )
        with open(os.path.join(project_dir, "project.yml"), "r") as f:
            project_yaml = yaml.safe_load(f)
        if not self.__verify_function(package_name, function_name, project_yaml):
            raise errors.ResourceError(
                f"No package with name {package_name} and fuction with {function_name} found in {os.path.join(project_dir, 'project.yml')}"
            )
        function_index = f"{package_name}/{function_name}"
        label = self.spec.get("label")
        if label is None:
            raise errors.ResourceError("No label found in DOFunction target specs")
        namespaces_list_command = "doctl serverless namespaces list -ojson"
        namespace_id = None
        try:
            p = resources.Resource.execute(namespaces_list_command, project_dir)
            namespaces = p.stdout.decode("UTF-8")
            namespaces_list = json.loads(namespaces)
            for namespace in namespaces_list:
                if namespace.get("label") == label:
                    namespace_id = namespace.get("namespace")
        except CalledProcessError as e:
            raise errors.ResourceError(e.stdout.decode("UTF-8"))
        if namespace_id is None:
            create_namespace_command = (
                f"doctl serverless namespaces create --label {label} --region {region}"
            )
            try:
                p = resources.Resource.execute(create_namespace_command, project_dir)
                namespace_id = p.stdout.decode("UTF-8").split(" ")[4]
            except CalledProcessError as e:
                raise errors.ResourceError(
                    f"Error creating namespace in DO: {e.stderr.decode('UTF-8')}"
                )
        try:
            resources.Resource.execute(
                f"doctl serverless connect {namespace_id}", project_dir
            )
        except CalledProcessError as e:
            raise errors.ResourceError(
                f"Error connecting to namespace: {e.stderr.decode('UTF-8')}"
            )
        deploy_command = f"doctl serverless deploy . -ojson"
        try:
            resources.Resource.execute(deploy_command, config.get("project_dir"))
        except CalledProcessError as e:
            raise errors.ResourceError(e.stdout.decode("UTF-8"))
        try:
            p = resources.Resource.execute(
                f"doctl sls fn get {function_index} --url", project_dir
            )
            return p.stdout.decode("UTF-8")
        except CalledProcessError as e:
            raise errors.ResourceError(
                f"Error when getting function URL: {function_index}, {e.stderr.decode('UTF-8')}"
            )


def get_plugin_manager():
    """
    Return the plugin manager for hash-targets plugins.
    """
    pm = pluggy.PluginManager("hash-targets")
    pm.add_hookspecs(TargetSpec)
    pm.load_setuptools_entrypoints("hash-targets")
    pm.register(FakeTarget(), "FakeTarget")
    pm.register(K8STarget(), "K8STarget")
    pm.register(DockerRegistryTarget(), "DockerRegistryTarget")
    pm.register(DOFunctionTarget(), "DOFunctionTarget")
    return pm


def get_target(target, args, space):
    """
    Docs Return a target by its name

    Args:
        target (str): The name of the target to return.
        args (dict): A dictionary which contains the config for this target

    Return:
        object: The target object, which can be used to run actions
    """
    pm = get_plugin_manager()
    plugins = pm.list_name_plugin()
    for plugin in plugins:
        if target == plugin[0]:
            hash_target = plugin[1]
            hash_target.init(args, space)
            return hash_target



"""
The main kern module, it has the main_action method which runs an action x on resource y in environment z
"""

from hash import errors
import os
import graphviz
from hash import kern
from hash.kern.execute import Execute
from hash.kern.planner import Planner
from hash.resources.base import get_resource, ResourceSpace


def main_action(
    resource_file: str,
    action_name: str,
    env_name: str,
    base_path: str,
    store,
    plan_file=None,
):
    if os.path.isdir(resource_file):
        resource_file = os.path.join(resource_file, "resource.yaml")
    if not os.path.exists(resource_file):
        raise errors.ResourceError(f"File not found {resource_file}")
    resource = get_resource(resource_file)
    if resource is None:
        raise errors.ResourceError(f"Resource at {resource_file} is not recognized")
    rs = ResourceSpace(base_path, store)
    if env_name is None:
        env = None
    else:
        env = rs.find_resource_by_id(f"Env:{env_name}")
        if env is None:
            raise errors.ResourceError(f"No env with name {env_name}")
    h = rs.calculate_hash(resource)
    st = store.get(resource.getId(), h)
    pl = Planner(store, base_path)
    pl.plan(action_name, resource, env, st.get_imdict())
    if plan_file is not None:
        edges = pl.getGraph().getEdges()
        dot = graphviz.Digraph(
            f"Plan to run {action_name} on {resource.getId()} in {env_name}"
        )
        for edge in edges:
            node1_key = " ".join(edge.node1.getKey().split(":")[0:4])
            node2_key = " ".join(edge.node2.getKey().split(":")[0:4])
            dot.edge(node1_key, node2_key)
        with open(plan_file, "w") as f:
            f.write(dot.source)
        print(f"Plan saved to {plan_file}")
        return
    deps = pl.get__deps()
    ex = Execute(rs, store, "SerialExecutor", pl.get_states())
    try:
        ex.execute(pl.getGraph(), deps)
    except Exception as e:
        raise e
    finally:
        store.persist_states()
        for ac in kern.all_actions:
            artifacts = st.get_artifacts(ac, env_name).get(ac, [])
            for artifact in artifacts:
                if artifact.getKind() == "file":
                    try:
                        os.remove(artifact.getData())
                    except Exception:
                        pass
        for _, state in pl.get_states().items():
            for ac in kern.all_actions:
                res = rs.find_resource_by_id(state.get_resource_id())
                if res:
                    env_name_res = res.getMetadata("env")
                    if env_name_res:
                        artifacts = state.get_artifacts(ac, env_name_res).get(ac, [])
                    else:
                        artifacts = state.get_artifacts(ac, env_name).get(ac, [])
                else:
                    artifacts = state.get_artifacts(ac, env_name).get(ac, [])
                for artifact in artifacts:
                    if artifact.getKind() == "file":
                        try:
                            os.remove(artifact.getData())
                        except Exception:
                            pass

    return ex.get_steps()



"""
This module is used to render templates for a resource and clear them, it needs resource, state and globals as input
"""

import os
import shlex
import subprocess
import jinja2

from hash.utils import Artifacts, Template, Globals

from hash import errors


def raise_resource_error(name, e):
    if name == "build":
        raise errors.ResourceBuildError(e)
    elif name == "test":
        raise errors.ResourceTestError(e)
    elif name == "publish":
        raise errors.ResourcePublishError(e)
    elif name == "deploy":
        raise errors.ResourceDeployError(e)
    else:
        raise errors.ResourceError(e)


class Renderer(object):
    def __init__(self, resource, space, store, globs, env=None) -> None:
        self.__resource = resource
        self.__space = space
        self.__store = store
        self.__globs = globs
        self.__env = env

    def __get_hash_templates(self, path, recursive) -> list:
        if recursive is True:
            find_command = f"find {path} -type f -name *.hash"
        else:
            find_command = f"find {path} -maxdepth 1 -type f -name *.hash"
        find_command = shlex.split(find_command)
        path = os.path.join(self.__space.get_base(), self.__resource.getPath())
        find = subprocess.run(find_command, cwd=path, stdout=subprocess.PIPE)
        return find.stdout.decode().split("\n")[:-1]

    def get_hash_templates(self):
        direct_templates = self.__get_hash_templates(".", False)
        templates = self.__resource.getSpec("templates", {})
        templates_files = []
        templates_dirs = []
        templates_dirs_loop = []
        if templates is not {}:
            templates_files = templates.get("file", [])
            template_dirs = templates.get("dirs", [])
            for dir in template_dirs:
                templates_dirs.extend(self.__get_hash_templates(dir, False))
            template_dirs_loop = templates.get("dirs_loop", [])
            for dir in template_dirs_loop:
                templates_dirs_loop.extend(self.__get_hash_templates(dir, True))

        return templates_dirs + templates_dirs_loop + templates_files + direct_templates

    def get_artifact_by_id(self, id: str, action: str, name: str) -> dict:
        resource = self.__space.find_resource_by_id(id)
        if resource is None:
            raise errors.ResourceError(f"No resource with id {id}, and artifact {name}")
        resource_hash = self.__space.calculate_hash(resource)
        if self.__store is None:
            raise errors.ResourceError("No available store to get artifacts")
        st = self.__store.get(id, resource_hash)
        if st is None:
            raise errors.ResourceError(f"No state for resource with id {id}")
        if self.__env is None:
            artifacts = st.get_artifacts(action, self.__env)
        elif type(self.__env) == str:  # TODO: This should not happen, investigate later
            artifacts = st.get_artifacts(action, self.__env)
        else:
            artifacts = st.get_artifacts(action, self.__env.getName())
        for artifact in artifacts[action]:
            if artifact.getId() == name:
                return artifact
        raise errors.ResourceError(
            f"No artifact with id {name} found in resource with id {id} and in env {self.__env} for action {action}"
        )

    def get_envs(self, env_name=None, dry=False) -> dict:
        envs = {}
        resource = self.__resource
        parent = resource.getMetadata("parent")
        parents = [resource]
        while parent:
            resource = self.__space.find_resource_by_id(parent)
            if resource:
                parents.append(resource)
                parent = resource.getMetadata("parent")
            else:
                break
        parents.reverse()
        for parent in parents:
            env = parent.getSpec("envs", {})
            envs.update(env)
        globals = Globals(self.__globs, env_name, dry)
        envs["globals"] = globals
        artifacts = Artifacts(dry, self.get_artifact_by_id)
        envs["artifacts"] = artifacts
        return envs

    def get_deps(self):
        """
        Return deps defined in templates
        """
        hash_files = self.get_hash_templates()
        path = os.path.join(self.__space.get_base(), self.__resource.getPath())
        tmp = Template(path, True)
        envs = self.get_envs(None, True)
        for file in hash_files:
            try:
                tmp.template(file, **envs)
            except jinja2.exceptions.UndefinedError as e:
                pass
            except errors.TemplateRenderError:
                pass
        return envs["globals"].deps + envs["artifacts"].deps

    def render(self, env=None, dry=False, action=None):
        hash_files = self.get_hash_templates()
        path = os.path.join(self.__space.get_base(), self.__resource.getPath())
        tmp = Template(path, dry)
        envs = self.get_envs(env, dry)
        for file in hash_files:
            try:
                ret = tmp.template(file, **envs)
            except jinja2.exceptions.UndefinedError as e:
                if not dry:
                    raise_resource_error(
                        action, f"error during rendering file {file}: {e}"
                    )
            if dry:
                return envs
            else:
                new_file = ".".join(file.split(".")[:-1])
                new_file = os.path.join(path, new_file)
                with open(new_file, "w") as f:
                    f.write(ret)

    def clear(self) -> None:
        hash_files = self.get_hash_templates()
        path = os.path.join(self.__space.get_base(), self.__resource.getPath())
        for file in hash_files:
            new_file = ".".join(file.split(".")[:-1])
            new_file = os.path.join(path, new_file)
            try:
                os.remove(new_file)
            except FileNotFoundError:
                pass



""""
This module contains the Action class which represents executes a single action assuming all its deps are ready
"""

import os
import subprocess
from hash import errors, utils
from hash.resources.base import EnvResource, Resource
from .state import all_actions


def is_action_less(action1: str, action2: str) -> bool:
    if action1 == "build":
        return True
    if action1 == "test" and (
        action2 == "deploy" or action2 == "publish" or action2 == "test"
    ):
        return True
    if action1 == "publish" and (action2 == "deploy" or action2 == "publish"):
        return True
    if action1 == "deploy" and action2 == "deploy":
        return True
    return False


class Action(object):
    """
    The Action class is used to execute a single action on a single resource with
    a state and en env, it assumes that all dependencies are ready and all templates
    are rendered for this resource.
    """

    def __init__(
        self,
        action: str,
        resource: Resource,
        state: utils.ImDict,
        env: EnvResource,
        base_path: str,
        reason: str,
    ) -> None:
        """
        Action class initializer, it takes 5 arguments which are needed to execute the action

        args:
            action (str): The action string that we want to run it can be build, test, publish or deploy.
            resource (Resource): The resource object that we want to execute the action on it.
            state (utils.ImDict): The immutable version of the state for this resource.
            env (EnvResource): The environment object in which this action will be run.
            base_path (str): The base path for resources on disk, this is needed for executing action scripts if found.
            reason (str): The reason for running this action
        """
        utils.check_action(action)
        self.__action = action
        utils.check_type(
            state,
            utils.ImDict,
            errors.ActionError,
            False,
            f"State must be an immutable dictionary: found {type(state)}",
        )
        self.__state = state
        self.__resource = resource
        utils.check_type(
            env,
            EnvResource,
            errors.ActionError,
            True,
            f"Env must be EnvResource: found {type(env)}",
        )
        self.__env = env
        self.__base_path = base_path
        self.__path = os.path.join(self.__base_path, self.__resource.getPath())
        self.__reason = reason

    def raise_error(self, msg):
        if self.__action == "build":
            raise errors.ResourceBuildError(msg)
        elif self.__action == "test":
            raise errors.ResourceTestError(msg)
        elif self.__action == "publish":
            raise errors.ResourcePublishError(msg)
        elif self.__action == "deploy":
            raise errors.ResourceDeployError(msg)

    def get_json(self):
        if self.__env:
            return {
                "action": self.__action,
                "resource_id": self.__resource.getId(),
                "env": self.__env.getId(),
            }
        return {
            "action": self.__action,
            "resource_id": self.__resource.getId(),
            "env": None,
        }

    def setup_env(self) -> None:
        """
        This method is used to setup the environment for running a command or script on the resource, it
        sets these enviornment variables:

        R_NAME: It contains the name of the resource.
        R_PARENT: It contains the ID of parent resource (Kind:Name)
        R_PARENT_NAME: It contains the name of parent resource
        R_ENV: It contains the name of environment where the action is being executed.
        R_ACTION: It contains the current name of action being executed.
        R_SPEC_{X}: this defines a group of environment variables for every spec element, for example
        if you have a spec called 'x' with a value of 'abc' it will create an environment variable
        called R_SPEC_X with a value of 'abc'
        """
        os.environ["R_NAME"] = self.__resource.getName()
        parentId = self.__resource.getParentId()
        if parentId is None:
            parentId = ""
            parentName = ""
        else:
            parentName = parentId.split(":")[1]
        if self.__env is None:
            env = ""
        else:
            env = self.__env.getName()
        os.environ["R_ENV"] = env
        os.environ["R_ACTION"] = self.__action
        os.environ["R_PARENT"] = parentId
        os.environ["R_PARENT_NAME"] = parentName
        for k, v in self.__resource.getSpecDict().items():
            env_name = f"R_SPEC_{k.upper()}"
            os.environ[env_name] = str(v)

    def re_action(self):
        """
        Check if we need to re-run the action based on the resource's specs
        """
        re_action_command = self.__resource.getSpec(f"re_{self.__action}_command")
        if re_action_command:
            self.setup_env()
            try:
                self.__resource.execute(re_action_command, self.__path)
            except subprocess.CalledProcessError as e:
                if e.returncode == 1:
                    return True
                else:
                    pass  # TDOD: the return code chould be 1, what to do here?
        else:
            re_action_script = self.__resource.getSpec(f"re_{self.__action}_script")
            if re_action_script:
                self.setup_env()
                script = utils.Script(os.path.join(self.__path, re_action_script))
                try:
                    script.run(self.__path)
                except subprocess.CalledProcessError as e:
                    if e.returncode == 1:
                        return True
                    else:
                        pass  # TDOD: the return code chould be 1, what to do here?
            else:
                if hasattr(self.__resource, "re_action") and callable(
                    getattr(self.__resource, "re_action")
                ):
                    if self.__resource.re_action(
                        self.__action, self.__state, self.__env
                    ):
                        return True
        return False

    def do_pre_action(self):
        """
        Run pre action command or script on the resource, it searches for a spec
        called `pre_<action>_command` if found then it is executed, if not it searches
        for `pre_<action>_script` if found then it is executed and if not then the method
        returns.

        Before executing any command or script the values for environment variables are
        set using setup_env method
        """
        pre_action_command = self.__resource.getSpec(f"pre_{self.__action}_command")
        if pre_action_command:
            self.setup_env()
            self.__resource.execute(pre_action_command, self.__path)
        else:
            pre_action_script = self.__resource.getSpec(f"pre_{self.__action}_script")
            if pre_action_script:
                self.setup_env()
                script = utils.Script(os.path.join(self.__path, pre_action_script))
                script.run(self.__path)

    def do_post_action(self):
        """
        Run post action command or script on the resource, it searches for a spec
        called `post_<action>_command` if found then it is executed, if not it searches
        for `post_<action>_script` if found then it is executed and if not then the method
        returns.

        Before executing any command or script the values for environment variables are
        set using setup_env method
        """
        post_action_command = self.__resource.getSpec(f"post_{self.__action}_command")
        if post_action_command:
            self.setup_env()
            self.__resource.execute(post_action_command, self.__path)
        else:
            post_action_script = self.__resource.getSpec(f"post_{self.__action}_script")
            if post_action_script:
                self.setup_env()
                script = utils.Script(os.path.join(self.__path, post_action_script))
                script.run(self.__path)

    def do_action(self):
        """
        Run the actual action on the resource, it searches for `<action>_command`
        spec if found then it is ran, if not it searches for `<action>_script` spec
        if found then it is ran otherwise it calls `action` method on the resource and
        passes to it the action's name, state and env, the returned value is used
        as a return for this method, it no return value was found then it returns
        a default dictionary with one `msg` key, it also checks if the return value
        is not a dictionary, it raises an error according to the current action.
        """
        action_command = self.__resource.getSpec(f"{self.__action}_command")
        action_script = self.__resource.getSpec(f"{self.__action}_script")
        result = None
        if action_command:
            self.setup_env()
            self.__resource.execute(action_command, self.__path)
        elif action_script:
            if action_script:
                self.setup_env()
                script = utils.Script(os.path.join(self.__path, action_script))
                script.run(self.__path)
        else:
            result = self.__resource.action(self.__action, self.__state, self.__env)
            if result is not None and type(result) != dict:
                self.raise_error(
                    f"Returned value from action must be dict: found {type(result)}"
                )
        if result is None:
            result = {"msg": f"Run {self.__action} without errors"}
        return result

    def run_action(self):
        """
        Run an action on a resource assuming all deps are ready and templates rendered

        It first calls pre action code, then calls the actual action and lastly the post action code.
        """
        if self.__action not in all_actions:
            raise errors.ActionError(
                f"Action {self.__action} not allowed: allowed actions are {all_actions}"
            )
        print(
            f"Run action {self.__action} on resource {self.__resource.getId()} in env {self.__env}, reason: {self.__reason}"
        )
        self.do_pre_action()
        ret = self.do_action()
        self.do_post_action()
        return ret



from setuptools import setup, find_packages


def requirements():
    with open("requirements.txt", "r") as f:
        return f.read().splitlines()


with open("README.md", "r") as fh:
    long_description = fh.read()
setup(
    name="hashkern",
    version="0.3.2",
    author="Mouhsen Ibrahim",
    author_email="mouhsen.ibrahim@gmail.com",
    description="A tool for managing resources in a mono repository",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Topic :: Software Development",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src", "tests": "tests"},
    python_requires=">=3.7",
    entry_points={"console_scripts": ["haks=hash.cli:main"]},
    packages=find_packages("src"),
    install_requires=requirements(),
    project_urls={
        "Documentation": "https://hash-kern.readthedocs.io/en/latest/index.html",
        "Source": "https://gitlab.com/hash-platform/hashkern",
    },
)



import argparse
import configparser
import os
import sys
from hash import errors

from hash.kern import main_action, __VERSION__
from hash.kern.templates import Renderer
from hash.store import get_store
from hash.resources import ResourceSpace, get_resource


def __read_config(args):
    config = configparser.ConfigParser()
    config.read(args.config)
    sections = config.sections()
    if sections == []:
        sys.exit(9)
    return config


def build(args):
    configs = __read_config(args)
    try:
        config = dict(configs[args.storage])
    except KeyError:
        config = {}
    try:
        s = get_store(args.storage, config)
    except errors.StoreNotFound as e:
        print(f"Error in storage plugin {args.storage}, {e}")
        sys.exit(1)
    s.init(config)
    return main_action(args.path, "build", args.env, os.environ["PWD"], s, args.plan)


def test(args):
    configs = __read_config(args)
    try:
        config = dict(configs[args.storage])
    except KeyError:
        config = {}
    try:
        s = get_store(args.storage, config)
    except errors.StoreNotFound as e:
        print(f"Error in storage plugin {args.storage}, {e}")
        sys.exit(1)
    s.init(config)
    return main_action(args.path, "test", args.env, os.environ["PWD"], s, args.plan)


def publish(args):
    configs = __read_config(args)
    try:
        config = dict(configs[args.storage])
    except KeyError:
        config = {}
    try:
        s = get_store(args.storage, config)
    except errors.StoreNotFound as e:
        print(f"Error in storage plugin {args.storage}, {e}")
        sys.exit(1)
    s.init(config)
    return main_action(args.path, "publish", args.env, os.environ["PWD"], s, args.plan)


def deploy(args):
    configs = __read_config(args)
    try:
        config = dict(configs[args.storage])
    except KeyError:
        config = {}
    try:
        s = get_store(args.storage, config)
    except errors.StoreNotFound as e:
        print(f"Error in storage plugin {args.storage}, {e}")
        sys.exit(1)
    s.init(config)
    return main_action(args.path, "deploy", args.env, os.environ["PWD"], s, args.plan)


def _hash(args):
    configs = __read_config(args)
    try:
        config = dict(configs[args.storage])
    except KeyError:
        config = {}
    try:
        s = get_store(args.storage, config)
    except errors.StoreNotFound as e:
        print(f"Error in storage plugin {args.storage}, {e}")
        sys.exit(1)
    s.init(config)
    rs = ResourceSpace(os.environ["PWD"], s)
    r = get_resource(args.path)
    return rs.calculate_hash(r)


def render(args):
    configs = __read_config(args)
    try:
        config = dict(configs[args.storage])
    except KeyError:
        config = {}
    try:
        s = get_store(args.storage, config)
    except errors.StoreNotFound as e:
        print(f"Error in storage plugin {args.storage}, {e}")
        sys.exit(1)
    s.init(config)
    rs = ResourceSpace(os.environ["PWD"], s)
    r = get_resource(args.path)
    r = Renderer(r, rs, s, s.get_globals(), args.env)
    r.render(args.env)


def clear(args):
    configs = __read_config(args)
    try:
        config = dict(configs[args.storage])
    except KeyError:
        config = {}
    try:
        s = get_store(args.storage, config)
    except errors.StoreNotFound as e:
        print(f"Error in storage plugin {args.storage}, {e}")
        sys.exit(1)
    s.init(config)
    rs = ResourceSpace(os.environ["PWD"], s)
    r = get_resource(args.path)
    r = Renderer(r, rs, s, {}, args.env)
    r.clear()


def version(args):
    return __VERSION__


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(
        prog="hash",
        description="A tool to build resources based on their hash and type",
    )
    sub_parsers = parser.add_subparsers(dest="subparser_name")
    parser.add_argument(
        "--storage",
        help="The storage system used default is Local File",
        default="LocalFile",
    )
    parser.add_argument(
        "--config",
        help="The configuration file default is config.ini",
        default="config.ini",
    )
    parser.add_argument(
        "--env", help="An environment to run the action in it", default=None
    )
    parser.add_argument(
        "--plan", help="Only do a plan and save it to this file", default=None
    )
    parser_build = sub_parsers.add_parser("build")
    parser_build.add_argument("path", help="path to build")
    parser_build.set_defaults(func=build)
    parser_test = sub_parsers.add_parser("test")
    parser_test.add_argument("path", help="path to test")
    parser_test.set_defaults(func=test)
    parser_publish = sub_parsers.add_parser("publish")
    parser_publish.add_argument("path", help="path to publish")
    parser_publish.set_defaults(func=publish)
    parser_deploy = sub_parsers.add_parser("deploy")
    parser_deploy.add_argument("path", help="path to deploy")
    parser_deploy.set_defaults(func=deploy)

    parser_hash = sub_parsers.add_parser("hash")
    parser_hash.add_argument("path", help="path to hash")
    parser_hash.set_defaults(func=_hash)

    parser_render = sub_parsers.add_parser("render")
    parser_render.add_argument("path", help="path to render")
    parser_render.set_defaults(func=render)

    parser_clear = sub_parsers.add_parser("clear")
    parser_clear.add_argument("path", help="path to clear")
    parser_clear.set_defaults(func=clear)

    parser_version = sub_parsers.add_parser("version")
    parser_version.set_defaults(func=version)

    args = parser.parse_args(argv)
    if args.subparser_name is None:
        parser.print_help()
        sys.exit(1)
    try:
        return args.func(args)
    except errors.ResourceError as e:
        return f"Error: {e}"


if __name__ == "__main__":
    output = main(sys.argv[1:])
    if output:
        print(output)



"""
The main DAG module, it contains definitions for nodes, edges and methods to walk the graph
"""

from .utils import check_type
from .errors import GraphError, NodeError, EdgeError
from hash import utils

class Node(object):
    """
    A graph node
    """
    def __init__(self, key, metadata) -> None:
        utils.check_type(key, str, NodeError, False, f"Node key must be a string: found {type(key)}")
        self.__key = key
        self.__metadata = metadata

    def __repr__(self) -> str:
        return f"<Node key: {self.__key}>"
    def __eq__(self, other: object) -> bool:
        if type(other) == str:
            return self.__key == other
        if type(other) == Node:
            return self.__key == other.getKey()
        raise NodeError(f"Cannot compare node with {type(other)}")
    def __hash__(self) -> int:
        return hash(self.__key)
    def getKey(self):
        return self.__key
    def getMetadata(self, key = None):
        if key:
            return self.__metadata.get(key)
        return self.__metadata
    def setMetadata(self, key, value):
        self.__metadata[key] = value
class Edge(object):
    """
    An edge between two nodes
    """
    def __init__(self, node1 : Node, node2 : Node) -> None:
        check_type(node1, Node, GraphError, False, f"Graph nodes should be of class hash.dag.Node : found {type(node1)}")
        check_type(node2, Node, GraphError, False, f"Graph nodes should be of class hash.dag.Node : found {type(node2)}")
        self.node1 = node1
        self.node2 = node2

    def __eq__(self, other: object) -> bool:
        if type(other) == Edge:
            return self.node1 == other.node1 and self.node2 == other.node2
        raise EdgeError(f"Cannot compare edge with {type(other)}")

    def __repr__(self) -> str:
        return f"<Edge node1: {self.node1}, node2: {self.node2}>"

class Graph(object):
    """
    A simple graph
    """
    def __init__(self) -> None:
        self.__nodes = []
        self.__edges = []

    def addEdge(self, edge : Edge):
        edge.node1 = self.addNode(edge.node1)
        edge.node2 = self.addNode(edge.node2)
        for e in self.__edges:
            if e == edge:
                return
        self.__edges.append(edge)
    def removeEdge(self, edge):
        for i in range(len(self.__edges)):
            if self.__edges[i] == edge:
                del self.__edges[i]
                return
    def addNode(self, node : Node):
        for n in self.__nodes:
            if n == node:
                return n
        self.__nodes.append(node)
        return node

    def upToEdges(self, node : Node) -> list:
        result = []
        for e in self.__edges:
            if e.node2 == node:
                result.append(e.node1)
        return result

    def downToEdges(self, node : Node) -> list:
        result = []
        for e in self.__edges:
            if e.node1 == node:
                result.append(e.node2)
        return result


    def check_loop(self, node):
        self.loop = False
        self.dfs_walk(node, self.__check_loop)
        return self.loop

    def __check_loop(self, node, neighbors, visited_nodes):
        for n in neighbors:
            if n in visited_nodes:
                self.loop = True
                return

    def dfs_walk(self, node : Node, callback):
        return self.__dfs_walk(node, callback, {})
    def __dfs_walk(self, node : Node, callback, visisted_nodes):
        if visisted_nodes.get(node.getKey()):
            return
        visisted_nodes[node.getKey()] = 1
        nodes = self.downToEdges(node)
        callback(node, nodes, visisted_nodes)
        for n in nodes:
            self.__dfs_walk(n, callback, visisted_nodes)

    def __walk_number(self, node, distance):
        if len(self.ns) == 0:
            node.label = 0
            self.ns[0] = [node]
        else:
            upToNodes = self.upToEdges(node)
            max_label = 0
            for n in upToNodes:
                try:
                    if n.label > max_label:
                        max_label = n.label
                except AttributeError:
                    return
            node.label = max_label + 1
            nodes = self.ns.get(node.label, [])
            if nodes:
                self.ns[node.label].append(node)
            else:
                self.ns[node.label] = [node]
    def __next_index(self, ns, index, max_num) -> int:
        for i in range(index + 1, max_num+1):
            try:
                ns[i]
                return i
            except KeyError:
                pass
    def walk_number(self):
        # Use kahn's algorithm
        l = [[]]
        s = []
        for n in self.__nodes:
            if len(self.upToEdges(n)) == 0:
                n.setMetadata("_index", 0)
                s.append(n)
        index = 1
        while len(s) != 0:
            node = s.pop(0)
            if node.getMetadata('_index') == index:
                index += 1
                l.append([])
            l[index-1].append(node)
            for n in self.downToEdges(node):
                self.removeEdge(Edge(node, n))
                if len(self.upToEdges(n)) == 0:
                    n.setMetadata("_index", index)
                    s.append(n)

        return l
    def bfs_walk_number(self, node):
        self.ns = {}
        self.bfs_walk(node, self.__walk_number)
        indices = self.ns.keys()
        num = max(indices)
        for i in range(num):
            try:
                self.ns[i]
            except KeyError:
                nextIndex = self.__next_index(self.ns, i, num)
                if nextIndex is None:
                    break
                self.ns[i] = self.ns[nextIndex]
                del self.ns[nextIndex]
        return self.ns

    def bfs_walk(self, node : Node, callback):
        q = [node]
        z = {
            node.getKey(): 1
        }
        e = {
            node.getKey(): 0
        }
        visited_nodes = []
        callback(node, 0)
        while len(q) != 0:
            n = q[0]
            del q[0]
            nodes = self.downToEdges(n)
            for nn in nodes:
                if z.get(nn.getKey()):
                    continue
                callback(nn, e[n.getKey()] + 1)
                try:
                    nn.label
                    z[nn.getKey()] = 1
                    e[nn.getKey()] = e[n.getKey()] + 1
                    q.append(nn)
                    visited_nodes.append(nn)
                except AttributeError:
                    continue

    def getNodes(self):
        return self.__nodes

    def getEdges(self):
        return self.__edges

    def __repr__(self) -> str:
        return f"Graph <edges: {self.__edges}>"
    def __eq__(self, __o: object) -> bool:
        if len(self.__nodes) != len(__o.__nodes):
            return False
        if len(self.__edges) != len(__o.__edges):
            return False
        for node in self.__nodes:
            upToNodes1 = self.upToEdges(node)
            upToNodes2 = __o.upToEdges(node)
            if len(upToNodes1) != len(upToNodes2):
                return False
            return upToNodes1 == upToNodes2



"""
Main hash module for calculating and storing hashes
"""

from dirhash import dirhash

from hash import logs

logger = logs.getLogger(__name__)


def setupLogger(output, level):
    logs.setOutput(logger, output)
    logs.setLevel(logger, level)


class Hash (object):
    """
    Docs for Hash class
    """
    __supported_algs = ["md5", "sha1", "sha256", "sha512"]

    def __init__(self, alg="sha256") -> None:
        if alg not in Hash.__supported_algs:
            raise ValueError(
                f"Algorith {alg} is not supported, please select from {Hash.__supported_algs}")
        self.alg = alg

    def hash(self, path, match=None, ignore=None) -> str:
        """
        hash method
        """
        if match:
            return dirhash(path, self.alg, match=match, ignore=ignore)
        return dirhash(path, self.alg, ignore=ignore)

    def __repr__(self) -> str:
        return f"<Hash alg: {self.alg}>"



"""
State module defines State class with all the methods, functions needed to manage state, it ensures
that the internal state structure is hidden from consumers of the module and that only methods
are used to manipulate or read state.
"""

from datetime import datetime
import json

from hash import errors, utils

all_actions = ["build", "test", "publish", "deploy"]
globals_key = "globals"
artifacts_key = "artifacts"


class State(object):
    """
    This class defines a single state for a single resource, hash and resource ID combination make the
    state unique.
    """

    def __init__(self, _hash: str, resource_id: str) -> None:
        if type(_hash) != str:
            raise errors.StateError(f"hash must be string not {type(_hash)}")
        if type(resource_id) != str:
            raise errors.StateError(
                f"resource id must be string not {type(resource_id)}"
            )
        self.__hash = _hash
        self.__resource_id = resource_id
        self.__envs = {}
        self.__no_env = {}
        self.__status = {"last_action": "", "error": ""}
        self.__deps = {}
        self.__vcs = {}
        self.__local = {"__no_env": {}}
        self.__updated_at = datetime.now()

    def __check_result_artifacts(self, result: dict, action: str, env=None):
        artifacts = result.get(artifacts_key)
        if artifacts is None:
            return
        utils.check_type(
            artifacts,
            list,
            errors.StateArtifactError,
            False,
            f"Artifacts must be a list: found {type(artifacts)}",
        )
        for index in range(len(artifacts)):
            if type(artifacts[index]) == dict:
                artifacts[index] = utils.Artifact.read_artifact(artifacts[index])
            utils.check_type(
                artifacts[index],
                utils.Artifact,
                errors.StateArtifactError,
                False,
                f"Every artifact must be of type utils.Artifact: found {type(artifacts[index])}",
            )
            if artifacts[index].getAction() != action:
                raise errors.StateArtifactError(
                    f"Found artifact for action {artifacts[index].getAction()} in {action} result"
                )
            if artifacts[index].getEnv() != env:
                raise errors.StateArtifactError(
                    f"Found artifact for env {artifacts[index].getEnv()} in {action} result of env {env}"
                )

    def set_result(self, result: dict, action: str, env=None):
        """
        This method is used to set the result of executing an action on the resource

        args:
            result (dict): The result object that we want to store in state.
            action (str): The action that generated this result.
            env (str|None): The env that the action was executed in, if None store the result in no_env

        raises:
            errors.StateError: When result, action or env types are wrong
        """
        utils.check_action_env_result(action, env, result)
        self.__check_result_artifacts(result, action, env)
        if env is None:
            self.__no_env[action] = result.copy()
        else:
            if self.__envs.get(env) is None:
                self.__envs[env] = {action: result.copy()}
            else:
                self.__envs[env][action] = result.copy()
        self.__updated_at = datetime.now()

    def get_result(self, action: str, env=None) -> utils.ImDict:
        """
        Retrieve a result immutable dictionary from the state.

        args:
            action (str): The action that we want to retrieve its result.
            env (str|None): The env where the action was executed.
        returns:
            utils.ImDict: An immutable dictionary of the result as it is stored in state.
        """
        utils.check_action_env(action, env)
        if env is None:
            return utils.ImDict(self.__no_env.get(action, {}))
        return utils.ImDict(self.__envs.get(env, {}).get(action, {}))

    def __get_result_key(self, action: str, key: str, env=None, default=None):
        """
        Get a key from the result of one or more actions
        """
        ret = {}
        if action == "":
            for ac in all_actions:
                if env is None:
                    ret[ac] = self.__no_env.get(ac, {}).get(key, default)
                else:
                    ret[ac] = self.__envs.get(env, {}).get(ac, {}).get(key, default)
        else:
            if env is None:
                ret[action] = self.__no_env.get(action, {}).get(key, default)
            else:
                ret[action] = self.__envs.get(env, {}).get(action, {}).get(key, default)
        if type(ret) == dict:
            return utils.ImDict(ret)
        elif type(ret) == list:  # TODO: Is this needed?
            return tuple(ret)
        return ret

    def set_result_local(self, result: dict, action: str, env=None):
        """
        This method is used to set the result of executing an action on the resource in local state

        args:
            result (dict): The result object that we want to store in state.
            action (str): The action that generated this result.
            env (str|None): The env that the action was executed in, if None store the result in no_env

        raises:
            errors.StateError: When result, action or env types are wrong
        """
        utils.check_action_env_result(action, env, result)
        if env is None:
            self.__local["__no_env"][action] = result.copy()
        else:
            if self.__local.get(env) is None:
                self.__local[env] = {action: result.copy()}
            else:
                self.__local[env][action] = result.copy()
        self.__updated_at = datetime.now()

    def get_result_local(self, action: str, env=None) -> utils.ImDict:
        """
        Retrieve a result immutable dictionary from the local state.

        args:
            action (str): The action that we want to retrieve its result.
            env (str|None): The env where the action was executed.
        returns:
            utils.ImDict: An immutable dictionary of the result as it is stored in state.
        """
        utils.check_action_env(action, env)
        if env is None:
            return utils.ImDict(self.__local["__no_env"].get(action, {}))
        return utils.ImDict(self.__local.get(env, {}).get(action, {}))

    def get_artifacts(self, action: str, env=None) -> utils.ImDict:
        """
        Retrieve the artifacts in this state.

        args:
            action (str): The action which we want to retrieve artifacts for it
            env (str|None): The env where the action was executed.

        returns:
            utils.ImDict or tuple : An immutable dictionary or tuple of the artifacts object
        """
        utils.check_action_env(action, env)
        return self.__get_result_key(action, artifacts_key, env, [])

    def get_globals(self, action: str, env=None) -> utils.ImDict:
        """
        Retrieve the global values in this state.

        args:
            action (str): The action which we want to retrieve globals for it
            env (str|None): The env where the action was executed.

        returns:
            utils.ImDict: An immutable dictionary of the globals object
        """
        utils.check_action_env(action, env)
        return self.__get_result_key(action, globals_key, env, {})

    def get_hash(self) -> str:
        """
        Return the hash of this state.

        returns:
            str: The hash value
        """
        return self.__hash

    def get_resource_id(self) -> str:
        """
        Return the resource id of this state

        returns:
            str: The id of the resource.
        """
        return self.__resource_id

    def get_envs(self) -> utils.ImDict:
        """
        Return the immutable version of envs dictionary

        returns:
            utils.ImDict: An immutable dictionary which contains all envs and their execution results
        """
        return utils.ImDict(self.__envs)

    def get_no_env(self) -> utils.ImDict:
        """
        Return the immutable version of no_env dictionary

        returns:
            utils.ImDict:  An immutable dictionary which contains the no_env execution results
        """
        return utils.ImDict(self.__no_env)

    def get_env_names(self, action: str) -> list:
        """
        Returns a list of env names which the action was executed on them, if the action is "" then return only all envs

        args:
            action (str): The name of the action

        returns:
            list (str): A list of strings which contain the env names
        """
        if action == "":
            return [key for key, _ in self.__envs.items()]
        else:
            return [
                key for key, item in self.__envs.items() if item.get(action) != None
            ]

    def add_dep(self, dep_id: str, dep_hash: str):
        """
        Add a new dependency with its hash

        args:
            dep_id (str): The id of the dependency to add
            dep_hash (str): The hash of the dependecny to add
        """
        utils.check_type(
            dep_id,
            str,
            errors.StateError,
            False,
            f"dep_id should be string: found {type(dep_id)}",
        )
        utils.check_type(
            dep_hash,
            str,
            errors.StateError,
            False,
            f"dep_hash should be string: found {type(dep_hash)}",
        )
        self.__deps[dep_id] = dep_hash

    def get_deps(self) -> utils.ImDict:
        """
        Return an immutable dictionary of all dependencies with their hashes

        returns:
            utils.ImDict: An immutable dictionary of all dependencies with their hashes
        """
        return utils.ImDict(self.__deps)

    def get_dep(self, dep_id: str) -> str:
        """
        Return the hash of a single dependency

        args:
            dep_id (str): The id of the dependency to return its hash

        returns:
            str: The hash of the dependency, None if the id does not exist
        """
        utils.check_type(
            dep_id,
            str,
            errors.StateError,
            False,
            f"dep_id should be string: found {type(dep_id)}",
        )
        return self.__deps.get(dep_id)

    def set_status(self, last_action: str, error: str):
        """
        Set the status of the state, it contains the last action that was ran without errors
        and the last action that caused an error if possible.

        args:
            last_action (str): The last action without an error
            error (str): The last action with an error
        """
        utils.check_type(
            last_action,
            str,
            errors.StateError,
            False,
            f"last action should be sring: found {type(last_action)}",
        )
        utils.check_type(
            error,
            str,
            errors.StateError,
            False,
            f"error should be sring: found {type(error)}",
        )
        if last_action != "" and last_action not in all_actions:
            raise errors.StateError(
                f"last_action must be a valid action found {last_action}"
            )
        if error != "" and error not in all_actions:
            raise errors.StateError(f"error must be a valid action found {error}")
        self.__status = {"last_action": last_action, "error": error}

    def get_status(self) -> utils.ImDict:
        """
        Return the status immutable dictionary of the state

        returns:
            utils.ImDict: An immutable dictionary of the status
        """
        return utils.ImDict(self.__status)

    def get_local(self) -> utils.ImDict:
        """
        Return the local state immutable dictionary of the state

        returns:
            utils.ImDict: An immutable dictionary of the local state
        """
        return utils.ImDict(self.__local)

    def set_local(self, local):
        """
        Set the local state

        args:
            local (dict): The local state to set
        """
        utils.check_type(
            local,
            dict,
            errors.StateError,
            False,
            f"Local state should be dict found : {type(local)}",
        )
        self.__local = local

    def get_json_str(self, return_id=False) -> str:
        """
        Return a string JSON representation of the state

        args:
            return_id (bool): A boolean value that controls if we should return resource ID in state or not

        returns:
            str: The string JSON representation
        """
        ret = {
            "hash": self.get_hash(),
            "deps": self.get_deps(),
            "envs": self.get_envs(),
            "no_env": self.get_no_env(),
            "status": self.get_status(),
            "local": self.get_local(),
            "__updated_at": self.__updated_at,
        }
        if return_id:
            ret["resource_id"] = self.get_resource_id()
        return json.dumps(ret, cls=utils.HashJsonEncoder)

    def get_json(self, return_id=False) -> dict:
        """
        Return a JSON representation of the state

        args:
            return_id (bool): A boolean value that controls if we should return resource ID in state or not

        returns:
            dict: The JSON representation
        """
        ret = {
            "hash": self.get_hash(),
            "deps": self.get_deps(),
            "envs": self.get_envs(),
            "no_env": self.get_no_env(),
            "status": self.get_status(),
            "__updated_at": self.__updated_at,
        }
        if return_id:
            ret["resource_id"] = self.get_resource_id()
        return ret

    def last_error(self) -> str:
        """
        Return the last action that caused an error, return "" if no error exists

        returns:
            str: The last action that caused an error
        """
        return self.__status.get("error", "")

    def last_action(self) -> str:
        """
        Return the last action exeucted without an error, if no action was executed return ""

        returns:
            str: The last action without an error
        """
        return self.__status.get("last_action", "")

    def copy_state(self, st):
        """
        A method to copy everything from another state object except for hash and resource ID

        args:
            st (State): The other state object to copy
        """
        self.__envs = st.get_envs().copy()
        self.__no_env = st.get_no_env().copy()
        self.__deps = st.get_deps().copy()
        self.__status = st.get_status().copy()
        self.__local = st.get_local().copy()
        self.__updated_at = self.get_updated_at().copy()

    def get_imdict(self) -> utils.ImDict:
        """
        Return a read only dictionary of the state.

        returns:
            utils.ImDict: An immutable dictionary of the entire state
        """
        return utils.ImDict(
            {
                "hash": self.__hash,
                "resource_id": self.__resource_id,
                "envs": self.__envs,
                "no_env": self.__no_env,
                "deps": self.__deps,
                "status": self.__status,
                "local": self.__local,
            }
        )

    @classmethod
    def read_state(cls, state: dict, resource_id=None):
        """
        A classmethod to convert a state dictionary into a valid state object.

        args:
            state (dict): The dictionary that we want to convert into a state
            resource_id (str): The ID of resource, it is used only if the ID is not stored in state.

        returns:
            State: A State object
        """
        utils.check_type(
            state,
            dict,
            errors.StateError,
            False,
            f"State must be a dictionary: found {type(state)}",
        )
        _resource_id = state.get("resource_id")
        if _resource_id is None:
            utils.check_type(
                resource_id,
                str,
                errors.StateError,
                False,
                f"Passed resource id must be string: found {type(resource_id)}",
            )
        else:
            utils.check_type(
                _resource_id,
                str,
                errors.StateError,
                False,
                f"State resource id must be string: found {type(resource_id)}",
            )
            resource_id = _resource_id
        st = cls(state.get("hash"), resource_id)
        status = state.get("status", {})
        utils.check_type(
            status,
            dict,
            errors.StateError,
            False,
            f"Status must be dictionary: found {type(status)}",
        )
        st.set_status(status.get("last_action", ""), status.get("error", ""))
        deps = state.get("deps", {})
        utils.check_type(
            deps,
            dict,
            errors.StateError,
            True,
            f"deps must be a list: found {type(deps)}",
        )
        for dep_id, dep_hash in deps.items():
            st.add_dep(dep_id, dep_hash)
        envs = state.get("envs", {})
        utils.check_type(
            envs,
            dict,
            errors.StateError,
            False,
            f"envs must be a dictionary: found {type(envs)}",
        )
        for env_name, v in envs.items():
            utils.check_type(
                env_name,
                str,
                errors.StateError,
                False,
                f"env name must be string: found {type(env_name)}",
            )
            utils.check_type(
                v,
                dict,
                errors.StateError,
                False,
                f"env actions must be a dictionary: found {type(v)}",
            )
            for ac_name, result in v.items():
                st.set_result(result, ac_name, env_name)
        no_env = state.get("no_env", {})
        utils.check_type(
            no_env,
            dict,
            errors.StateError,
            False,
            f"no_env must be a dictionary: found {type(envs)}",
        )
        for ac_name, result in no_env.items():
            st.set_result(result, ac_name)
        local = state.get("local", {"__no_env": {}})
        for env_name, v in local.items():
            if env_name == "__no_env":
                env_name = None
            else:
                utils.check_type(
                    env_name,
                    str,
                    errors.StateError,
                    False,
                    f"env name must be string: found {type(env_name)}",
                )
            utils.check_type(
                v,
                dict,
                errors.StateError,
                False,
                f"env actions must be a dictionary: found {type(v)}",
            )
            for ac_name, result in v.items():
                st.set_result_local(result, ac_name, env_name)
        return st
